from typing import Iterable, Self, Tuple, Hashable
from collections.abc import MutableSet
from abc import ABC, abstractmethod

class Symbol:
    """
    Ex:
    ===
        >>> A = Symbol("Some Name", 2)
        >>> A(3, 4) # this is a fact

    """

    def __init__(self, name: str, arity: int):
        self.arity = arity
        self.name = name

    def __call__(self, *args) -> "Atom":
        assert len(args) == self.arity
        return Fact(self, *args)

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name and self.arity == other.arity

    def __hash__(self):
        return hash((self.arity, self.name))


class v(str):
    """
    Ex:
    ===
        >>> x = v("x")
        >>> x2 = v("x")
        >>> x == x2
        True
    """

    def __eq__(self, other):
        return type(self) is type(other) and super().__eq__(other)

    def __repr__(self):
        return f"v({self})"


class Atom:
    """
    Ex:
    ===
        >>> A = Symbol("Some Name", 3)
        >>> Atom(A, (3, 4, v("x")))
        A(3, 4, v("x"))
        
    """
    def __init__(self, symbol: Symbol, data: tuple): 
        self.symbol = symbol
        self.data = data

    def __hash__(self):
        return hash((self.symbol, self.data))


class Fact(Atom):
    def __init__(self, symbol: Symbol, data: tuple): 
        super().__init__(symbol, data)
        # TODO


class CQ:
    """
    Ex:
    ===
        # Q(x, x, y) := Exists z R(x, y, z)
        >>> R = Symbol("R", 3)
        >>> CQ((v("x"), v("x"), v("y")), (R(v("x"), v("y"), v("z")),))
    """
    def __init__(self, out_variables: tuple[v], *atoms: tuple[Atom, ...]):
        self.out_variables = out_variables
        self.atoms = atomes

    @property
    def variables(self) -> frozenset[v]:
        # TODO

    @property
    def is_boolean(self):
        # TODO
        
    def containment(self, query: CQ) -> bool:
        # TODO


class UCQ(frozenset[CQ]): ...

class StaticDatabase(frozenset[Fact]):
    @property
    def relations(self): 
        #TODO

    def naive_evaluation(self, query: CQ | UCQ) -> set(tuple):
        # TODO

    @property
    def active_domain(self) -> frozenset:
        #TODO

    def __repr__(self):
        return f"DB({super().__repr__()})"



class DynamicDatabase(MutableSet[Fact]):
    def __init__(self):
        self.datastore : set[fact] = set()

    @property
    def relations(self): 
        #TODO

    def naive_evaluation(self, query: CQ | UCQ) -> set(tuple):
        # TODO

    @property
    def active_domain(self) -> frozenset:
        #TODO
